class Plotter {
  /**
   *
   * @param {*} data = [x,y1,y2,...]
   * @param {*} parent
   */
  constructor(
    data,
    parent,
    options = {
      alarmData: [],
      alarmColor: (value) => {
        return value == 1 ? "#FE0000" : value == 2 ? "#FFA8FE" : null;
      },
      colors: [],
      header: [],
      clickCallback: null,
      levelerCallback: null,
    }
  ) {
    this.colors = options.colors;
    this.header = options.header;
    this.clickCallback = options.clickCallback;
    this.levelerCallback = options.levelerCallback;
    // this.header = data[0];
    this.xData = data.map((d) => d[0]);
    this.numCols = data[0].length - 1;
    this.yData = [];
    for (let i = 0; i < this.numCols; i++)
      this.yData.push(data.map((d) => d[i + 1]));
    this.data = data;
    this.parent = parent;
    // this.options = { header: options.header, colors: options.colors };

    /// construct essential d3 components
    this.widthOuter = parent.width();
    this.heightOuter = parent.height();
    this.marginOuter = { top: 10, right: 30, bottom: 20, left: 60 };
    // this.distMainSub = 40;
    this.widthInner =
      this.widthOuter - this.marginOuter.left - this.marginOuter.right;
    this.heightOuterSub = Math.floor(this.heightOuter / 5);
    this.heightOuterMain = this.heightOuter - this.heightOuterSub;
    this.heightInnerMain =
      this.heightOuterMain - this.marginOuter.top - this.marginOuter.bottom;
    this.heightInnerSub =
      this.heightOuterSub - this.marginOuter.top - this.marginOuter.bottom;

    this.xMain = d3
      .scaleTime()
      .range([0, this.widthInner])
      .domain(d3.extent(this.xData));

    this.xz = this.xMain;
    this.xSub = d3
      .scaleTime()
      .range([0, this.widthInner])
      .domain(this.xMain.domain());
    this.xMainAxis = d3.axisBottom(this.xMain);

    this.xSubAxis = d3.axisBottom(this.xSub);
    this.yMain = [];
    this.yMainAxes = [];
    this.ySub = [];
    this.ySubAxes = [];
    this.paths = [];

    // alarms
    // plot background
    this.yBckg = d3
      .scaleLinear()
      .domain([0, 1])
      .range([this.heightInnerMain, 0]);
    this.yBckgSub = d3
      .scaleLinear()
      .domain([0, 1])
      .range([this.heightInnerSub, 0]);

    let a1Data = [];
    for (let i = 0; i < options.alarmData.length; i++) {
      if (
        options.alarmData[i][1] == 1 ||
        (i > 0 && options.alarmData[i - 1][1] == 1)
      ) {
        a1Data.push([options.alarmData[i][0], 1]);
      } else {
        a1Data.push([options.alarmData[i][0], undefined]);
      }
    }

    this.a1 = {
      color: options.alarmColor(1),
      data: a1Data,
    };

    a1Data = [];
    for (let i = 0; i < options.alarmData.length; i++) {
      if (
        options.alarmData[i][1] == 2 ||
        (i > 0 && options.alarmData[i - 1][1] == 2)
      ) {
        a1Data.push([options.alarmData[i][0], 1]);
      } else {
        a1Data.push([options.alarmData[i][0], undefined]);
      }
    }

    this.a2 = {
      color: options.alarmColor(2),
      data: a1Data,
    };

    this.areaA = d3
      .area()
      .defined((d) => !isNaN(d[1]))
      .x((d) => this.xMain(d[0]))
      .y0(this.yBckg(0))
      .y1((d) => this.yBckg(d[1]));

    this.areaB = d3
      .area()
      .defined((d) => !isNaN(d[1]))
      .x((d) => this.xSub(d[0]))
      .y0(this.yBckgSub(0))
      .y1((d) => this.yBckgSub(d[1]));

    this.zoom = d3
      .zoom()
      .scaleExtent([1, Infinity])
      .extent([
        [0, 0],
        [this.widthInner, this.heightInnerMain],
      ])
      .translateExtent([
        [0, 0],
        [this.widthInner, this.heightInnerMain],
      ])
      .on("zoom", (event) => {
        // console.log(event);
        if (!event.sourceEvent) return;
        const xz = event.transform.rescaleX(this.xMain);
        this.a1Path.attr("d", this.getArea(this.a1, xz));
        this.a2Path.attr("d", this.getArea(this.a2, xz));
        this.paths.forEach((path, ind) =>
          path.attr("d", this.getLine(ind, xz))
        );
        this.xz = xz;
        this.gx.call(d3.axisBottom(xz));
        this.gb.call(
          this.brush.move,
          xz.domain().map((el) => this.xSub(el))
        );
      });
  }
  getArea(al, x) {
    return d3
      .area()
      .defined((d) => !isNaN(d[1]))
      .x((d) => x(d[0]))
      .y0(this.yBckg(0))
      .y1((d) => this.yBckg(d[1]))(al.data);
  }
  getLine(ind, x) {
    return d3
      .line()
      .x((d) => x(d[0]))
      .y((d) => this.yMain[ind](d[this.cols[ind]]))(this.data);
  }

  plot(
    cols,
    options = {
      colors: [],
      header: [],
    }
  ) {
    if (cols.length < 1) return;
    this.cols = cols;
    let colHeight = Math.floor(
      (this.heightInnerMain - (cols.length - 1) * 10) / cols.length
    );
    // let colHeightSub = Math.floor(
    //   (this.heightInnerSub - (cols.length - 1) * 10) / cols.length
    // );

    $("#theGraph").remove();
    this.parent.append(`
    <div class="row" id="theGraph"><svg></svg></div>
    `);
    const svg = d3
      .select("svg")
      .attr("viewBox", [0, 0, this.widthOuter, this.heightOuterMain])
      .append("g")
      .attr(
        "transform",
        "translate(" + this.marginOuter.left + "," + this.marginOuter.top + ")"
      )
      .attr("id", "transformMainSvg")
      .on("click", (event) => {
        let [x, y] = d3.pointer(event, d3.select("#transformMainSvg").node()); //d3.select("#transformMainSvg")
        if (typeof this.clickCallback === "function")
          this.clickCallback(
            this.xz.invert(x),
            this.yMain.map((yz) =>
              yz.domain()[0] <= yz.invert(y) && yz.domain()[1] >= yz.invert(y)
                ? yz.invert(y)
                : null
            )
          );
      });

    $("#theGraph").append(`<svg id="brushSvg"></svg>`);
    const svgBrush = d3
      .select("#brushSvg")
      .attr("viewBox", [0, 0, this.widthOuter, this.heightOuterSub])
      .append("g")
      .attr(
        "transform",
        "translate(" + this.marginOuter.left + "," + this.marginOuter.top + ")"
      );

    this.gx = svg
      .append("g")
      .attr("transform", `translate(0,${this.heightInnerMain})`)
      .call(this.xMainAxis);
    this.gb = svgBrush
      .append("g")
      .attr("transform", `translate(0,${this.heightInnerSub})`)
      .call(this.xSubAxis);

    svg
      .append("clipPath")
      .attr("id", "clipRect")
      .append("rect")
      .attr("x", 0)
      .attr("y", 0)
      .attr("width", this.widthInner)
      .attr("height", this.heightInnerMain);

    this.a1Path = svg
      .append("path")
      .datum(this.a1.data)
      .attr("clip-path", `url(#clipRect)`)
      .attr("fill", this.a1.color)
      .attr("d", this.areaA);
    this.a2Path = svg
      .append("path")
      .datum(this.a2.data)
      .attr("clip-path", `url(#clipRect)`)
      .attr("fill", this.a2.color)
      .attr("d", this.areaA);

    svgBrush
      .append("path")
      .datum(this.a1.data)
      .attr("fill", this.a1.color)
      .attr("d", this.areaB);
    svgBrush
      .append("path")
      .datum(this.a2.data)
      .attr("fill", this.a2.color)
      .attr("d", this.areaB);
    // lines
    this.linesMain = [];
    this.linesSub = [];

    // prepare yAxes
    for (let i = 0; i < cols.length; i++) {
      let yStart = i * colHeight + i * 10;
      let yEnd = (i + 1) * colHeight + i * 10;
      let yStartSub = (yStart / this.heightInnerMain) * this.heightInnerSub; //i * colHeightSub + i * 10;
      let yEndSub = (yEnd / this.heightInnerMain) * this.heightInnerSub; //(i + 1) * colHeightSub + i * 10;
      // console.log(yStart, yEnd);

      let ind = i;
      let amount =
        0.05 *
        (d3.max(this.data, (d) => d[cols[ind]]) -
          d3.min(this.data, (d) => d[cols[ind]]));
      this.yMain.push(
        d3
          .scaleLinear()
          .range([yEnd, yStart])
          .domain([
            d3.min(this.data, (d) => d[cols[ind]]) - amount,
            d3.max(this.data, (d) => d[cols[ind]]) + amount,
          ])
      );
      this.yMainAxes.push(
        d3.axisLeft(this.yMain[ind].nice()).ticks(Math.max(3, colHeight / 15))
      );
      svg.append("g").call(this.yMainAxes[ind]);
      this.ySub.push(
        d3
          .scaleLinear()
          .range([yEndSub, yStartSub])
          .domain([
            d3.min(this.data, (d) => d[cols[ind]]) - amount,
            d3.max(this.data, (d) => d[cols[ind]]) + amount,
          ])
          .nice()
      );
      // this.ySubAxes.push(d3.axisLeft(this.ySub[ind]));

      this.linesMain.push(
        d3
          .line()
          .x((d) => this.xMain(d[0]))
          .y((d) => this.yMain[ind](d[cols[ind]]))
      );
      this.linesSub.push(
        d3
          .line()
          .x((d) => this.xSub(d[0]))
          .y((d) => this.ySub[ind](d[cols[ind]]))
      );

      let color =
        options.colors[ind] ||
        this.colors[cols[ind] - 1] ||
        "#" + Math.floor(Math.random() * 16777215).toString(16);
      this.paths.push(
        svg
          .append("path")
          .datum(this.data)
          .attr("clip-path", `url(#clipRect)`)
          .attr("class", "line")
          .attr("stroke", color)
          .attr("d", this.linesMain[ind])
      );
      svgBrush
        .append("path")
        .datum(this.data)
        .attr("class", "line")
        .attr("stroke", color)
        .attr("d", this.linesSub[ind]);
    }
    // for zooming
    svg.call(this.zoom);
    svg
      .append("rect")
      .attr("width", this.widthInner)
      .attr("height", this.heightInnerMain)
      .style("fill", "none")
      .style("pointer-events", "all");

    // brushing

    ////////////

    // brush
    this.brush = d3
      .brushX()
      .extent([
        [0, 0],
        [this.widthInner, this.heightInnerSub],
      ])
      .on("brush", (event) => {
        // console.log(event);
        if (!event.sourceEvent) return;
        if (event.selection) {
          const xz = this.xMain
            .copy()
            .domain(event.selection.map(this.xMain.invert, this.xMain));
          this.a1Path.attr("d", this.getArea(this.a1, xz));
          this.a2Path.attr("d", this.getArea(this.a2, xz));
          this.paths.forEach((path, ind) =>
            path.attr("d", this.getLine(ind, xz))
          );
          this.gx.call(d3.axisBottom(xz));
          this.xz = xz;
          svg.call(
            this.zoom.transform,
            d3.zoomIdentity
              .scale(
                this.widthInner / (event.selection[1] - event.selection[0])
              )
              .translate(-event.selection[0], 0)
          );
        }
      });
    // .on("end", (event) => {
    //   if (!event.selection) {
    //     this.gb.call(this.brush.move, defaultSelection);
    //   }
    // });
    // const defaultSelection = this.xSub.range();
    // console.log(defaultSelection);
    this.gb = svgBrush.append("g").attr("class", "brush").call(this.brush);
    // .call(this.brush.move, defaultSelection);

    // const symb = d3.symbol().type(d3.symbolTriangle);
    // draw leveler
    // const arrowSize = 30;
    // const arrow = svg
    //   .append("g")
    //   .append("path")
    //   .attr(
    //     "d",
    //     (d) =>
    //       `M 0 ${-arrowSize / 2} L ${arrowSize / 2} ${-arrowSize} L  ${
    //         arrowSize / 2
    //       } 0 Z`
    //   )
    //   // .style("opacity", "0.5")
    //   .style("fill", "black");
    this.leveler = svg
      .append("rect")
      .attr("x", 0)
      .attr("y", -this.marginOuter.top)
      .attr("width", this.marginOuter.top)
      .attr("height", this.heightInnerMain + this.marginOuter.top + 10)
      .style("fill", "black")
      .style("opacity", "0.5")
      .call(
        d3
          .drag()
          .on("drag", (event) => {
            if (event.x < 0 || event.x > this.widthInner) return;
            d3.select(this.leveler.node()).attr(
              "transform",
              "translate(" + event.x + "," + 0 + ")"
            );
          })
          .on("end", (event) => {
            // console.log(event);
            if (event.x < 0 || event.x > this.widthInner) return;
            let xs = this.xz.invert(event.x);
            let bs = 0;
            if (event.x > event.subject.x) bs = d3.bisectLeft(this.xData, xs);
            else bs = d3.bisectLeft(this.xData, xs) - 1;
            // d3.bisectCenter(this.xData, xs);
            let newX = this.xz(this.xData[bs]);
            if (newX < 0) {
              bs += 1;
              newX = this.xz(this.xData[bs]);
            }
            if (newX > this.widthInner) {
              bs -= 1;
              newX = this.xz(this.xData[bs]);
            }

            d3.select(this.leveler.node()).attr(
              "transform",
              "translate(" + newX + "," + 0 + ")"
            );
            this.computeYsForLevelerPos(bs);
          })
      );
    this.computeYsForLevelerPos(0);
    // svgBrush
    //   .append("rect")
    //   .attr("width", this.widthInner)
    //   .attr("height", this.heightInnerSub)
    //   .style("fill", "none")
    //   .style("pointer-events", "all");
  }

  computeYsForLevelerPos(bs) {
    if (typeof this.levelerCallback === "function")
      this.levelerCallback(
        this.xData[bs],
        this.cols.map(
          (col) => this.yData[col - 1][bs]
          // (((yz[bs + 1] || yz[bs]) - yz[bs]) /
          //   (this.xData[bs + 1] - this.xData[bs])) *
          //   (xs - this.xData[bs]) +
          // yz[bs]
        )
      );
  }

  getDomain() {
    return this.xz.domain();
  }

  plotAll() {
    for (let c = 3; c < this.header.length; c++) {
      this.createPlotForCol(c);
    }
  }
  createPlotForCol(col) {
    if ($("#svgRow" + col).length) {
      $("#svgRow" + col).remove();
    }
    this.parent.append(`
    <div class="row" id="svgRow${col}">
        <svg id="svg${col}" />
    </div>
    `);
    // set the dimensions and margins of the graph
    let margin = { top: 10, right: 30, bottom: 30, left: 60 },
      width = this.parent.width() - margin.left - margin.right,
      height = this.parent.height() - margin.top - margin.bottom;
    // console.log(margin);
    // append the svg object to the body of the page
    let svg = d3
      .select("#svg" + col)
      .attr(
        "viewBox",
        "0 0 " +
          (width + margin.left + margin.right) +
          " " +
          (height + margin.top + margin.bottom)
      )
      // .classed("svg-content", true)
      // .attr("width", width + margin.left + margin.right)
      // .attr("height", height + margin.top + margin.bottom)
      .append("g")
      .attr("transform", "translate(" + margin.left + "," + margin.top + ")");

    //Read the data

    // Now I can use this dataset:
    // Add X axis --> it is a date format
    let x = d3
      .scaleTime()
      .domain(d3.extent(this.data, (d) => d[0]))
      .range([0, width]);
    svg
      .append("g")
      .attr("transform", "translate(0," + height + ")")
      .call(d3.axisBottom(x));

    console.log(
      "y scale",
      d3.min(this.data, (d) => d[col]),
      d3.max(this.data, (d) => d[col])
    );
    // Add Y axis
    let amount =
      0.05 *
      (d3.max(this.data, (d) => d[col]) - d3.min(this.data, (d) => d[col]));
    let y = d3
      .scaleLinear()
      .domain([
        d3.min(this.data, (d) => d[col]) - amount,
        d3.max(this.data, (d) => d[col]) + amount,
      ])
      .range([height, 0]);
    svg.append("g").call(d3.axisLeft(y));

    // Add the line
    svg
      .append("path")
      .datum(this.data)
      .attr("fill", "none")
      .attr("stroke", "steelblue")
      .attr("stroke-width", 1.5)
      .attr(
        "d",
        d3
          .line()
          .x((d) => x(d[0]))
          .y((d) => y(d[col]))
      );
  }
  brushed() {}
  zoomed() {}
}
